package com.cmss.pretendProxy;

/*
 Java FTP Proxy Server 1.3.0
 Copyright (C) 1998-2014 Christian Schmidt

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Find the latest version at http://aggemam.dk/ftpproxy
 */

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.BindException;
import java.net.ConnectException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.List;

import javax.net.ServerSocketFactory;
import javax.net.SocketFactory;

public class FtpProxyServer extends ProxyServer {

    public FtpProxyServer() {

        super.logPrefix = "[FTP代理]";
    }

    /**
     * 初始化
     */
    protected void init() throws Exception {

        _socketFactory_ = SocketFactory.getDefault();
        _serverSocketFactory_ = ServerSocketFactory.getDefault();
    }

    @Override
    public void run() {

        infoLog("代理服务初始化");
        try {
            init();
        } catch (final Exception e) {
            errorLog("初始化代理服务失败", e);
            return;
        }

        infoLog("代理服务启动");

        final int bindPort = config.bindPort;
        ServerSocket ssControlClient = null;
        try {
            debugLog("代理服务启动开始" + cuttingLine);
            if (config.bindAddress == null) {
                ssControlClient = new ServerSocket(bindPort);
            } else {
                ssControlClient = new ServerSocket(bindPort, config.bindBacklog, config.bindAddress);
            }
            infoLog("代理服务器：" + ssControlClient.toString());
            debugLog("代理服务已启动" + cuttingLine);
            while (true) {
                final Socket skControlClient = ssControlClient.accept();
                infoLog("新连接：" + skControlClient.toString());
                _startProxyClient_(skControlClient);
            }
        } catch (final Exception e) {
            errorLog("启动代理服务出错", e);
        } finally {
            close(ssControlClient, "FTP代理服务");
            //代理出现异常  强制重启
            infoLog("代理服务已关闭" + cuttingLine);
            try {
                //异常后睡眠10秒重启
                Thread.sleep(10000);
            } catch (final InterruptedException e) {
                errorLog("", e);
            }
            infoLog("代理服务开始重启" + cuttingLine);
            final FtpProxyServer proxyServer = new FtpProxyServer();
            proxyServer.setConfig(config);
            proxyServer.start();
            infoLog("代理服务已重启" + cuttingLine);
        }

    }

    protected void _startProxyClient_(final Socket skControlClient) {

        new FtpProxy(skControlClient).start();

    }

    class FtpProxy extends Thread {

        Socket skControlClient, skControlServer;

        BufferedReader brClient, brServer;

        PrintStream psClient, psServer;

        ServerSocket ssDataClient, ssDataServer;

        Socket skDataClient, skDataServer;

        DataConnect dcData;

        // 接口的IP分别面向客户端和服务器端.
        String sLocalClientIP;

        String sLocalServerIP;

        //连接服务端模式  默认主动模式
        boolean serverPassive = false;

        //用户登录状态
        boolean userLoggedIn = false;

        boolean connectionClosed = false;

        public FtpProxy(final Socket skControlClient) {

            this.skControlClient = skControlClient;
        }

        public boolean isInSubnetList(final List list, final InetAddress ia) {

            if (list == null) {
                return false;
            }

            for (final Iterator iterator = list.iterator(); iterator.hasNext();) {
                final Subnet subnet = (Subnet) iterator.next();

                if (subnet.isInSubnet(ia)) {
                    return true;
                }
            }
            return false;
        }

        public int parsePort(String s) throws IOException {

            int port;
            try {
                int i = s.lastIndexOf('(');
                int j = s.lastIndexOf(')');
                if ((i != -1) && (j != -1) && (i < j)) {
                    s = s.substring(i + 1, j);
                }

                i = s.lastIndexOf(',');
                j = s.lastIndexOf(',', i - 1);

                port = Integer.parseInt(s.substring(i + 1));
                port += 256 * Integer.parseInt(s.substring(j + 1, i));
            } catch (final Exception e) {
                throw new IOException();
            }
            return port;
        }

        public synchronized ServerSocket getServerSocket(final int[] portRanges, final InetAddress ia)
                throws IOException {

            ServerSocket ss = null;
            if (portRanges != null) {
                // Current index of portRanges array.
                int i;
                int port;

                final Integer lastPort = (Integer) lastPorts.get(portRanges);
                if (lastPort != null) {
                    port = lastPort.intValue();
                    for (i = 0; i < portRanges.length && port > portRanges[i + 1]; i += 2) {
                        ;
                    }
                    port++;
                } else {
                    port = portRanges[0];
                    i = 0;
                }

                for (int lastTry = -2; port != lastTry; port++) {
                    if (port > portRanges[i + 1]) {
                        i = (i + 2) % portRanges.length;
                        port = portRanges[i];
                    }
                    if (lastTry == -1) {
                        lastTry = port;
                    }
                    try {
                        ss = new ServerSocket(port, 1, ia);
                        lastPorts.put(portRanges, new Integer(port));
                        break;
                    } catch (final BindException e) {
                        // Port already in use.
                    }
                }
            } else {
                ss = new ServerSocket(0, 1, ia);
            }
            return ss;
        }

        protected String formatLogPrefix(final String param) {

            String sourceAddress = null;
            String targetAddress = null;
            if (client2proxy.equals(param)) {
                sourceAddress = skControlClient.getInetAddress().getHostAddress() + ":" + skControlClient.getPort();
                targetAddress = skControlClient.getLocalAddress().getHostAddress() + ":"
                        + skControlClient.getLocalPort();
            } else if (proxy2client.equals(param)) {
                sourceAddress = skControlClient.getLocalAddress().getHostAddress() + ":"
                        + skControlClient.getLocalPort();
                targetAddress = skControlClient.getInetAddress().getHostAddress() + ":" + skControlClient.getPort();
            } else if (server2proxy.equals(param)) {
                sourceAddress = skControlServer.getInetAddress().getHostAddress() + ":" + skControlServer.getPort();
                targetAddress = skControlServer.getLocalAddress().getHostAddress() + ":"
                        + skControlServer.getLocalPort();
            } else if (proxy2server.equals(param)) {
                sourceAddress = skControlServer.getLocalAddress().getHostAddress() + ":"
                        + skControlServer.getLocalPort();
                targetAddress = skControlServer.getInetAddress().getHostAddress() + ":" + skControlServer.getPort();
            } else if (server2client.equals(param)) {
                sourceAddress = skControlServer.getInetAddress().getHostAddress() + ":" + skControlServer.getPort();
                targetAddress = skControlClient.getInetAddress().getHostAddress() + ":" + skControlClient.getPort();
            } else if (client2server.equals(param)) {
                sourceAddress = skControlClient.getInetAddress().getHostAddress() + ":" + skControlClient.getPort();
                targetAddress = skControlServer.getInetAddress().getHostAddress() + ":" + skControlServer.getPort();
            }
            return "[" + sourceAddress + "--->" + targetAddress + "]" + "[" + param + "]";

        }

        @Override
        public void run() {

            try {

                brClient = new BufferedReader(new InputStreamReader(skControlClient.getInputStream()));
                psClient = new PrintStream(skControlClient.getOutputStream(), true);

                if ((config.allowFrom != null && !isInSubnetList(config.allowFrom, skControlClient.getInetAddress()))
                        || isInSubnetList(config.denyFrom, skControlClient.getInetAddress())) {
                    infoLog("该ip被限制连接：" + skControlClient.getInetAddress());
                    final String toClient = config.msgOriginAccessDenied;
                    psClient.print(toClient + CRLF);
                    debugLog(formatLogPrefix(proxy2client) + toClient);
                    skControlClient.close();
                    return;
                }

                debugLog("解析masquerade_host");
                try {
                    if (config.masqueradeHostname == null) {
                        sLocalClientIP = skControlClient.getLocalAddress().getHostAddress().replace('.', ',');
                    } else {
                        sLocalClientIP = InetAddress.getByName(config.masqueradeHostname.trim()).getHostAddress()
                                .replace('.', ',');
                    }
                } catch (final UnknownHostException e) {
                    errorLog("解析masquerade_host失败", e);
                    final String toClient = config.msgMasqHostDNSError;
                    psClient.print(toClient + CRLF);
                    debugLog(proxy2client + toClient);
                    skControlClient.close();
                    return;
                }

                final String toClient = config.msgConnect;
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);

                for (;;) {
                    final String s = brClient.readLine();
                    if (s == null) {
                        break;
                    }
                    readCommandFromClient(s);
                    // Exit if connection closed (response == 221,421,530).
                    if (connectionClosed) {
                        break;
                    }
                }

            } catch (final Exception e) {
                errorLog("", e);
                final String toClient = config.msgInternalError;
                psClient.print(toClient + CRLF);
                debugLog(proxy2client + toClient);
            } finally {
                if (!config.clientOneBindPort) {
                    close(ssDataClient, "ssDataClient");
                }
                if (!config.serverOneBindPort) {
                    close(ssDataServer, "ssDataServer");
                }
                close(skDataClient, "skDataClient");
                close(skDataServer, "skDataServer");
                close(psClient, "psClient");
                close(psServer, "psServer");
                close(dcData, "dcData");
                close(skControlClient, "skControlClient");
            }
        }

        /**
         * 从客户端读取命令
         *
         * @param fromClient
         * @throws IOException
         */
        private void readCommandFromClient(final String fromClient) throws IOException {

            final String cmd = fromClient.toUpperCase();

            if (!userLoggedIn && (cmd.startsWith("PASV") || cmd.startsWith("PORT"))) {
                debugLog(formatLogPrefix(client2proxy) + fromClient);
                // 不登录不支持模式选择.
                psClient.print("530 Not logged in." + CRLF);
                debugLog(formatLogPrefix(proxy2client) + fromClient);
                return;
            }

            if (cmd.startsWith("USER")) {
                //代理连接FTP服务器
                connectServer(fromClient);
            } else if (cmd.startsWith("PASS")) {
                //输入密码
                imputPassword(fromClient);
            } else if (cmd.startsWith("PASV") || cmd.startsWith("EPSV")) {
                //客户端被动
                setupDataConnectionUsePassive(fromClient);
            } else if (cmd.startsWith("PORT")) {
                //主动模式
                setupDataConnectionUseActive(fromClient);
            } else if (cmd.startsWith("NOOP")) {
                //空操作  响应有可能超时
                diposeNoop(fromClient);
            } else {
                psServer.print(fromClient + CRLF);
                debugLog(formatLogPrefix(client2server) + fromClient);
                readResponseFromServer(true);
            }

        }

        /**
         * 连接FTP服务器
         */
        private void connectServer(final String cmd) throws IOException {

            debugLog(formatLogPrefix(client2proxy) + cmd);

            String username = null;
            String hostname = null;
            int serverport = 21;

            final String userString = cmd.substring(5);
            // String userString = "ftp1@10.139.10.176:21";

            final int a = userString.lastIndexOf('@');
            int c = userString.lastIndexOf(':');
            if (c < a) {
                c = -1;
            }

            //网络URL 未实现
            //网络url使用 格式 "ftp1*10.139.10.176*21"

            if (c == -1) {
                username = userString.substring(0, a);
                hostname = userString.substring(a + 1);
            } else {
                username = userString.substring(0, a);
                hostname = userString.substring(a + 1, c);
                serverport = Integer.parseInt(userString.substring(c + 1));
            }
            debugLog("username:" + username);
            debugLog("hostname:" + hostname);
            debugLog("serverport:" + serverport);

            if (hostname == null) {
                final String toClient = config.msgIncorrectSyntax;
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
                skControlClient.close();
                return;
            }

            final InetAddress serverAddress = InetAddress.getByName(hostname);

            if ((config.allowTo != null && !isInSubnetList(config.allowTo, serverAddress))
                    || isInSubnetList(config.denyTo, serverAddress)) {
                infoLog("该ip被限制连接：" + serverAddress);
                final String toClient = config.msgDestinationAccessDenied;
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
                skControlClient.close();
                return;
            }

            debugLog("连接----> " + hostname + ":" + serverport);

            try {

                skControlServer = new Socket(serverAddress, serverport);
                brServer = new BufferedReader(new InputStreamReader(skControlServer.getInputStream()));
                psServer = new PrintStream(skControlServer.getOutputStream(), true);
                final String fromServer = brServer.readLine();
                debugLog(formatLogPrefix(server2proxy) + fromServer);
                if (fromServer.startsWith("421")) {
                    final String toClient = fromServer;
                    psClient.print(toClient + CRLF);
                    debugLog(formatLogPrefix(proxy2client) + toClient);
                    return;
                }
                skControlServer = _connectServer_(skControlServer, hostname, serverport);
            } catch (final ConnectException e) {
                errorLog("连接" + hostname + ":" + serverport + "失败", e);
                final String toClient = config.msgConnectionRefused;
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
                return;
            }

            brServer = new BufferedReader(new InputStreamReader(skControlServer.getInputStream()));
            psServer = new PrintStream(skControlServer.getOutputStream(), true);

            //服务端数据连接ip
            sLocalServerIP = skControlServer.getLocalAddress().getHostAddress().replace('.', ',');

            //登录
            final String toServer = "USER " + username;
            psServer.print(toServer + CRLF);
            debugLog(formatLogPrefix(proxy2server) + toServer);
            readResponseFromServer(true);
        }

        /**
         * 为扩展ssl用
         *
         * @param serverport
         * @param hostname
         *
         * @param skControlServer2
         */
        protected Socket _connectServer_(final Socket socket, final String hostname, final int serverport)
                throws IOException {

            return socket;
        }

        /**
         * 输入密码
         */
        private void imputPassword(final String cmd) throws IOException {

            psServer.print(cmd + CRLF);
            debugLog(formatLogPrefix(client2server) + "PASS *******");
            final String fromServer = brServer.readLine();
            psClient.print(fromServer + CRLF);
            debugLog(formatLogPrefix(server2client) + fromServer);

            //登录判断
            final int response = Integer.parseInt(fromServer.substring(0, 3));
            if (response == 230) {
                userLoggedIn = true;
            }

            _initServer_();
        }

        /**
         * 初始化一些参数
         */
        protected void _initServer_() throws IOException {

        }

        /**
         * 数据连接 主动模式
         */
        private void setupDataConnectionUseActive(final String cmd) throws IOException {

            //暂时未实现

            debugLog(formatLogPrefix(client2proxy) + cmd);

            if (ssDataClient != null && !config.clientOneBindPort) {
                try {
                    ssDataClient.close();
                } catch (final IOException ioe) {
                }
                ssDataClient = null;
            }
            if (skDataClient != null) {
                try {
                    skDataClient.close();
                } catch (final IOException ioe) {
                }
            }
            if (ssDataServer != null && !config.serverOneBindPort) {
                try {
                    ssDataServer.close();
                } catch (final IOException ioe) {
                }
            }
            if (skDataServer != null) {
                try {
                    skDataServer.close();
                } catch (final IOException ioe) {
                }
            }

            if (dcData != null) {
                dcData.close();
            }

            debugLog(formatLogPrefix(client2proxy) + cmd);

            try {
                //客户端主动
                final int clientPort = parsePort(cmd);
                skDataClient = new Socket(skControlClient.getInetAddress(), clientPort);

                String toClient = "200 PORT command successful.";
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);

                //服务端主动

                if (ssDataServer == null || !config.serverOneBindPort) {
                    //开启数据监听端口
                    ssDataServer = getServerSocket(config.serverBindPorts, skControlServer.getLocalAddress());
                }

                if (ssDataServer != null) {
                    final int serverPort = ssDataServer.getLocalPort();
                    final String toServer = "PORT " + sLocalServerIP + ',' + serverPort / 256 + ','
                            + (serverPort % 256);
                    //通知服务端来连
                    psServer.print(toServer + CRLF);
                    debugLog(formatLogPrefix(proxy2server) + toServer);
                    debugLog("端口号：" + serverPort);

                    readResponseFromServer(false);

                } else {
                    //无可用端口
                    toClient = "425 Cannot allocate local port.";
                    psClient.print(toClient + CRLF);
                    debugLog(formatLogPrefix(proxy2client) + toClient);
                }

            } catch (final IOException e) {
                final String toClient = "425 PORT command failed - try using PASV instead.";
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
                return;
            }

        }

        /**
         * 被动模式
         */
        private void setupDataConnectionUsePassive(final String cmd) throws IOException {

            debugLog(formatLogPrefix(client2proxy) + cmd);

            //创建新连接时先关闭旧连接
            if (ssDataClient != null && !config.clientOneBindPort) {
                try {
                    ssDataClient.close();
                } catch (final IOException ioe) {
                }
            }
            if (skDataClient != null) {
                try {
                    skDataClient.close();
                } catch (final IOException ioe) {
                }
            }
            if (ssDataServer != null && !config.serverOneBindPort) {
                try {
                    ssDataServer.close();
                } catch (final IOException ioe) {
                }
            }
            if (skDataServer != null) {
                try {
                    skDataServer.close();
                } catch (final IOException ioe) {
                }
            }
            if (dcData != null) {
                dcData.close();
            }

            if (ssDataClient == null || !config.clientOneBindPort) {
                ssDataClient = getServerSocket(config.clientBindPorts, skControlClient.getLocalAddress());
            }

            if (ssDataClient != null) {
                final int clientPort = ssDataClient.getLocalPort();
                //客户端被动
                String toClient;
                if (cmd.startsWith("EPSV")) {
                    //IPv6
                    toClient = "229 Entering Extended Passive Mode (|||" + clientPort + "|)";
                } else {
                    toClient = "227 Entering Passive Mode (" + sLocalClientIP + "," + clientPort / 256 + ","
                            + (clientPort % 256) + ")";
                }
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
                debugLog("端口号：" + clientPort);

                skDataClient = ssDataClient.accept();
                if (!config.clientOneBindPort) {
                    close(ssDataClient, "ssDataClient");
                }

                //服务端被动 IPV6 未处理
                final String toServer = cmd;
                psServer.print(toServer + CRLF);
                debugLog(formatLogPrefix(proxy2server) + toServer);

                //获取服务端返回的端口
                final String fromServer = readResponseFromServer(false);
                final int serverPort = parsePort(fromServer);
                debugLog("Server: " + skControlServer.getInetAddress() + ":" + serverPort);

                //与服务端建立连接
                skDataServer = _socketFactory_.createSocket();
                skDataServer.connect(new InetSocketAddress(skControlServer.getInetAddress(), serverPort));

            } else {
                final String toClient = "425 Cannot allocate local port..";
                psClient.print(toClient + CRLF);
                debugLog(formatLogPrefix(proxy2client) + toClient);
            }

        }

        /**
         * NOOP命令单独处理 可能无响应
         *
         * @param fromClient
         * @throws SocketException
         */
        private void diposeNoop(final String fromClient) throws SocketException {

            psServer.print(fromClient + CRLF);
            debugLog(formatLogPrefix(client2server) + fromClient);

            final int soTimeOut = skControlClient.getSoTimeout();
            skControlClient.setSoTimeout(3000);

            try {
                readResponseFromServer(true);
            } catch (final IOException e) {

            } finally {
                skControlClient.setSoTimeout(soTimeOut);
            }

        }

        /**
         * 从服务端读取响应
         *
         * @param true 直接推送给客户端 false 代理拦截
         * @return
         * @throws IOException
         */
        private String readResponseFromServer(final boolean forwardToClient) throws IOException {

            String fromServer = brServer.readLine();
            String firstLine = fromServer;
            final int response = Integer.parseInt(fromServer.substring(0, 3));
            if (fromServer.charAt(3) == '-') {
                final String multiLine = fromServer.substring(0, 3) + ' ';
                while (!fromServer.startsWith(multiLine)) {
                    if (forwardToClient) {
                        psClient.print(fromServer + CRLF);
                    }
                    debugLog(formatLogPrefix((forwardToClient ? server2client : server2proxy)) + fromServer);
                    fromServer = brServer.readLine();
                }
            }

            if (forwardToClient || response == 110) {
                psClient.print(fromServer + CRLF);
            }
            debugLog(formatLogPrefix((forwardToClient ? server2client : server2proxy)) + fromServer);

            // Check for successful login.
            if (response == 230) {
                userLoggedIn = true;
            } else if (response == 221 || response == 421 || response == 530) {
                if (userLoggedIn) {
                    connectionClosed = true;
                }
                userLoggedIn = false;
            }

            if (response == 150) {
                _openConection_();
            }

            if (response >= 100 && response <= 199) {
                firstLine = readResponseFromServer(true);
            }

            if (brServer.ready()) {
                firstLine = readResponseFromServer(true);
            }

            return firstLine;
        }

        /**
         * 开启数据连接
         *
         * @throws IOException
         */
        protected void _openConection_() throws IOException {

            (dcData = new DataConnect(skDataClient, skDataServer)).start();

        }

        public class DataConnect extends Thread implements Closeable {

            private final byte buffer[] = new byte[DATABUFFERSIZE];

            private final Socket[] sockets;

            private boolean isInitialized;

            private final Object mutex = new Object();

            // Each argument may be either a Socket or a ServerSocket.
            public DataConnect(final Socket s1, final Socket s2) {

                this.sockets = new Socket[] { s1, s2 };
            }

            @Override
            public void run() {

                BufferedInputStream bis = null;
                BufferedOutputStream bos = null;

                try {
                    // n = 0 - Thread Copy socket 0 to socket 1
                    // n = 1 - Thread Copy socket 1 to socket 0
                    final int n = isInitialized ? 1 : 0;
                    if (!isInitialized) {
                        isInitialized = true;
                        // In some cases thread socket[0] -> socket[1] thread can
                        // finish before socket[1] -> socket[0] has a chance to
                        // start,
                        // so synchronize on a semaphore
                        synchronized (mutex) {
                            new Thread(this).start();
                            try {
                                mutex.wait();
                            } catch (final InterruptedException e) {
                                // Never occurs.
                            }
                        }

                    }

                    bis = new BufferedInputStream(sockets[n].getInputStream());
                    bos = new BufferedOutputStream(sockets[1 - n].getOutputStream());

                    synchronized (mutex) {
                        mutex.notify();
                    }

                    for (;;) {
                        for (int i; (i = bis.read(buffer, 0, DATABUFFERSIZE)) != -1;) {
                            bos.write(buffer, 0, i);
                        }
                        break;
                    }
                    bos.flush();
                } catch (final SocketException e) {
                    // Socket closed.
                } catch (final IOException e) {
                    errorLog("数据连接异常", e);
                }
                close();
            }

            @Override
            public void close() {

                FtpProxyServer.this.close(sockets[0], "数据连接1");
                FtpProxyServer.this.close(sockets[1], "数据连接2");

            }
        }

    }

}
