package com.cmss.pretendProxy;

/*
 Java FTP Proxy Server 1.3.0
 Copyright (C) 1998-2014 Christian Schmidt

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 Find the latest version at http://aggemam.dk/ftpproxy
 */

import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.security.KeyStore;
import java.security.SecureRandom;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

import org.apache.commons.net.ftp.FTPSServerSocketFactory;
import org.apache.commons.net.ftp.FTPSSocketFactory;
import org.apache.commons.net.util.TrustManagerUtils;

public class FtpsProxyServer extends FtpProxyServer {

    private SSLContext context;

    public FtpsProxyServer() {

        super.logPrefix = "[FTPS代理]";
    }

    /**
     * 初始化context
     */
    @Override
    protected void init() throws Exception {

        final String keyName = "zhou_ks";
        final char[] keyPwd = "123456".toCharArray();
        final KeyStore keyStore = KeyStore.getInstance(KeyStore.getDefaultType());

        // 装载当前目录下的key store. 可用jdk中的keytool工具生成keystore
        InputStream in = null;
        keyStore.load(in = FtpsProxyServer.class.getClassLoader().getResourceAsStream(keyName), keyPwd);
        in.close();

        // 初始化key manager factory
        final KeyManagerFactory kmf = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
        kmf.init(keyStore, keyPwd);

        // 初始化ssl context
        final TrustManager trustManager = TrustManagerUtils.getValidateServerCertificateTrustManager();
        //            context = SSLContextUtils.createSSLContext("TLS", kmf.getKeyManagers(), trustManager);
        final SSLContext context = SSLContext.getInstance("TLS");
        context.init(kmf.getKeyManagers(), new TrustManager[] { trustManager }, new SecureRandom());

        this.context = context;
        this._socketFactory_ = new FTPSSocketFactory(context);
        this._serverSocketFactory_ = new FTPSServerSocketFactory(context);
    }

    @Override
    protected void _startProxyClient_(final Socket skControlClient) {

        new FtpsProxy(skControlClient).start();

    }

    class FtpsProxy extends FtpProxy {

        public FtpsProxy(final Socket skControlClient) {

            super(skControlClient);
        }

        @Override
        protected Socket _connectServer_(final Socket socket, final String hostname, final int serverport)
                throws IOException {

            final String toServer = "AUTH TLS";
            psServer.print(toServer + CRLF);
            debugLog(formatLogPrefix(proxy2server) + toServer);
            final String fromServer = brServer.readLine();
            //234 Proceed with negotiation
            debugLog(formatLogPrefix(server2proxy) + fromServer);
            final SSLSocketFactory ssf = context.getSocketFactory();
            final SSLSocket sslSocket = (SSLSocket) ssf.createSocket(skControlServer, hostname, serverport, false);
            sslSocket.setEnableSessionCreation(true);
            sslSocket.setUseClientMode(true);
            sslSocket.setNeedClientAuth(false);
            sslSocket.setWantClientAuth(false);
            sslSocket.startHandshake();
            return sslSocket;
        }

        @Override
        protected void _initServer_() throws IOException {

            //登录后相关ftps特有设置设置  与客户端无关
            //客户端连接优化后可去掉
            final String toServer = "PROT P";
            psServer.print(toServer + CRLF);
            debugLog(formatLogPrefix(proxy2server) + toServer);
            final String fromServer = brServer.readLine();
            debugLog(formatLogPrefix(server2proxy) + fromServer);
        }

        @Override
        protected void _openConection_() throws IOException {

            //被动模式
            final SSLSocket sslSocket = (SSLSocket) skDataServer;
            sslSocket.setEnableSessionCreation(true);
            sslSocket.setUseClientMode(true);
            sslSocket.setNeedClientAuth(false);
            sslSocket.setWantClientAuth(false);
            sslSocket.startHandshake();
            infoLog("握手成功");
            (dcData = new DataConnect(skDataClient, skDataServer)).start();

        }

    }

}
