/*------------------------------------------------------------------------------
 *******************************************************************************
 * COPYRIGHT China Mobile (SuZhou) Software Technology Co.,Ltd. 2017
 *******************************************************************************
 *----------------------------------------------------------------------------*/
package com.cmss.pretendProxy;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.Logger;

/**
 * 代理的启动类，主要工作，读取配置文件，启动代理服务器
 */
public class PretendProxy {

    private static final Logger LOGGER = Logger.getLogger(PretendProxy.class);

    private final static String ftpConfigFile = "ftpProxy.conf";

    private final static String ftpsConfigFile = "ftpsProxy.conf";

    private final static String cuttingLine = "==============================";

    public PretendProxy() {

    }

    /**
     * 代理启动
     */
    public void startProxy() throws Exception {

        infoLog("代理启动");
        infoLog("ftp代理服务启动");
        startProxy(ftpConfigFile, new FtpProxyServer());
        infoLog("ftps代理服务启动");
        startProxy(ftpsConfigFile, new FtpsProxyServer());

    }

    private int startProxy(String configFile, final ProxyServer proxyServer) throws Exception {

        infoLog("读取配置文件");
        debugLog("读取配置文件开始" + cuttingLine);
        final Properties properties = new Properties();
        configFile = PretendProxy.class.getClassLoader().getResource("").getPath().replaceAll("%20", " ") + configFile;
        infoLog("配置文件路径：" + configFile);
        try {
            properties.load(new FileInputStream(configFile));
        } catch (final IOException e) {
            errorLog("读取配置文件出错", e);
            throw e;
        }
        debugLog("读取配置文件结束" + cuttingLine);

        infoLog("解析配置文件");
        debugLog("解析配置文件开始" + cuttingLine);
        Configuration config;
        try {
            config = new Configuration(properties);
        } catch (final Exception e) {
            errorLog("解析配置文件出错", e);
            throw e;
        }
        // 在配置类解析时，配置类会删除配置变量——任何剩余的变量都是未知的，因此是无效的。
        if (properties.size() > 0) {
            errorLog("无效的配置变量" + properties.propertyNames().nextElement(), null);
            return -1;
        }
        debugLog("解析配置文件结束" + cuttingLine);

        //启动代理
        infoLog("启动代理");
        debugLog("启动代理开始" + cuttingLine);
        proxyServer.setConfig(config);
        proxyServer.start();
        debugLog("启动代理结束" + cuttingLine);
        return 0;
    }

    /**
     * 调试日志
     */
    private void debugLog(final String msg) {

        LOGGER.debug(msg);

    }

    /**
     * 普通日志
     */
    private void infoLog(final String msg) {

        LOGGER.info(msg);

    }

    /**
     * 错误日志
     */
    private void errorLog(final String msg, final Exception e) {

        LOGGER.error(msg, e);

    }

    public static void main(final String[] args) throws Exception {

        new PretendProxy().startProxy();

    }

}
